#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.interfaces base module

This module defines base package interfaces.
"""

from zope.annotation import IAttributeAnnotatable
from zope.container.constraints import contains
from zope.interface import Attribute, Interface, implementer
from zope.interface.interfaces import IObjectEvent, ObjectEvent
from zope.schema import Bool, Choice, List, Object, TextLine

from pyams_mail.interfaces import MAILERS_VOCABULARY_NAME
from pyams_scheduler.interfaces.task import IBaseTaskScheduling, ITask, ITaskHistory, \
    TASK_SCHEDULING_MODES_VOCABULARY
from pyams_security.interfaces import IContentRoles
from pyams_security.schema import PrincipalsSetField
from pyams_utils.interfaces import ZODB_CONNECTIONS_VOCABULARY_NAME
from pyams_zmq.interfaces import IZMQProcess

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


#
# Scheduler events
#

MANAGE_SCHEDULER_PERMISSION = 'pyams.ManageScheduler'
'''Permission used to manage tasks scheduler properties'''

MANAGE_TASKS_PERMISSION = 'pyams.ManageSchedulerTasks'
'''Permission used to manager scheduler tasks'''

SCHEDULER_MANAGER_ROLE = 'pyams.SchedulerManager'
'''Scheduler manager role'''

TASKS_MANAGER_ROLE = 'pyams.TasksManager'
'''Tasks scheduler manager role'''


class IBeforeRunJobEvent(IObjectEvent):
    """Interface for events notified before a job is run"""


@implementer(IBeforeRunJobEvent)
class BeforeRunJobEvent(ObjectEvent):
    """Before run job event"""


class IAfterRunJobEvent(IObjectEvent):
    """Interface for events notified after a job is run"""

    status = Attribute("Job execution status")

    result = Attribute("Job execution result")


@implementer(IAfterRunJobEvent)
class AfterRunJobEvent(ObjectEvent):
    """After run job event"""

    def __init__(self, obj, status, result):
        super().__init__(obj)
        self.status = status
        self.result = result


#
# Scheduler interface
#

SCHEDULER_NAME = 'Tasks scheduler'
SCHEDULER_STARTER_KEY = 'pyams_scheduler.start_handler'
SCHEDULER_HANDLER_KEY = 'pyams_scheduler.tcp_handler'
SCHEDULER_AUTH_KEY = 'pyams_scheduler.allow_auth'
SCHEDULER_CLIENTS_KEY = 'pyams_scheduler.allow_clients'

SCHEDULER_JOBSTORE_KEY = 'pyams_scheduler.jobs'


class ISchedulerProcess(IZMQProcess):
    """Scheduler process marker interface"""


class ISchedulerHandler(Interface):
    """Scheduler manager marker interface"""


TASKS_SCHEDULER_LABEL = _("Tasks scheduler")


class IScheduler(IAttributeAnnotatable):
    """Scheduler interface"""

    contains(ITask)

    zodb_name = Choice(title=_("ZODB connection name"),
                       description=_("Name of ZODB defining scheduler connection"),
                       required=False,
                       default='',
                       vocabulary=ZODB_CONNECTIONS_VOCABULARY_NAME)

    report_mailer = Choice(title=_("Reports mailer"),
                           description=_("Mail delivery utility used to send mails"),
                           required=False,
                           vocabulary=MAILERS_VOCABULARY_NAME)

    report_source = TextLine(title=_("Reports source"),
                             description=_("Mail address from which reports will be sent"),
                             required=False)

    notified_host = TextLine(title=_("Notified host"),
                             description=_("If websockets notifications are enabled, this is "
                                           "the host (including protocol) which will "
                                           "be notified"),
                             required=False)

    show_home_menu = Bool(title=_("Access menu from home"),
                          description=_("If 'yes', a menu will be displayed to get access to "
                                        "tasks scheduler from site admin home page"),
                          required=True,
                          default=False)

    internal_id = Attribute("Internal ID")

    def get_socket(self):
        """Get ZMQ socket matching scheduler utility"""

    def get_task(self, task_id):
        """Get task matching given task ID"""

    def get_jobs(self):
        """Get text output of running jobs"""

    tasks = List(title=_("Scheduler tasks"),
                 description=_("List of tasks assigned to this scheduler"),
                 value_type=Object(schema=ITask),
                 readonly=True)

    history = List(title=_("History"),
                   description=_("Task history"),
                   value_type=Object(schema=ITaskHistory),
                   readonly=True)


class ISchedulerRoles(IContentRoles):
    """Scheduler roles"""

    scheduler_managers = PrincipalsSetField(title=_("Scheduler managers"),
                                            role_id=SCHEDULER_MANAGER_ROLE,
                                            required=False)

    tasks_managers = PrincipalsSetField(title=_("Tasks manager"),
                                        role_id=TASKS_MANAGER_ROLE,
                                        required=False)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.task.handler.ftp module

This module a directory synchronization handler  using FTP protocol.
"""

import os
import stat
import time
from contextlib import contextmanager
from ftplib import Error, FTP
from tempfile import SpooledTemporaryFile
from time import mktime

from pyams_scheduler.interfaces.task.sync import DirectoryReadError, DirectoryWriteError, \
    IDirectoryHandler, \
    IDirectoryInfo
from pyams_scheduler.task.handler import BaseDirectoryHandler
from pyams_utils.adapter import adapter_config


__docformat__ = 'restructuredtext'


class FTPHelper:
    """Base FTP helper class"""

    def __init__(self, ftp, name, mode):
        self.ftp = ftp
        self.name = name
        self.mode = mode
        self.file = SpooledTemporaryFile(64 * 1024)

    def close(self):
        """Close temporary file"""
        self.file.close()


class FTPReader(FTPHelper):
    """FTP reader helper class"""

    def __init__(self, ftp, name, mode):
        super().__init__(ftp, name, mode)
        try:
            self.ftp.ftp_client.retrbinary('RETR {}'.format(self.ftp.get_path(self.name)),
                                           self.file.write)
        except Error as exc:
            raise DirectoryReadError from exc
        self.file.seek(0)

    def read(self, bufsize):
        """Read data from source"""
        return self.file.read(bufsize)


class FTPWriter(FTPHelper):
    """FTP writer helper class"""

    def write(self, data):
        """Write data to target"""
        self.file.write(data)

    def close(self):
        self.file.seek(0)
        try:
            self.ftp.ftp_client.storbinary('STOR {}'.format(self.ftp.get_path(self.name)),
                                           self.file)
        except Error as exc:
            raise DirectoryWriteError from exc
        self.file.close()


@adapter_config(name='ftp',
                required=IDirectoryInfo,
                provides=IDirectoryHandler)
class FTPDirectoryHandler(BaseDirectoryHandler):
    """FTP directory handler"""

    ftp_client = None

    def connect(self):
        directory_info = IDirectoryInfo(self.context)
        self.ftp_client = FTP()
        self.ftp_client.connect(directory_info.host[1],  # pylint: disable=unsubscriptable-object
                                directory_info.port)
        if directory_info.username:
            self.ftp_client.login(directory_info.username, directory_info.password)

    def close(self):
        self.ftp_client.quit()

    @contextmanager
    def open(self, name, mode='r'):
        if mode.startswith('r'):
            helper = FTPReader(self, self.get_path(name), mode)
        else:
            helper = FTPWriter(self, self.get_path(name), name)
        try:
            yield helper
        finally:
            helper.close()

    def mkdir(self, dirname, mode=0o500):
        self.ftp_client.mkd(self.get_path(dirname))

    def listdir(self):
        yield from sorted(filter(lambda name: name not in ('.', '..'),
                                 self.ftp_client.nlst(self.directory)))

    @staticmethod
    def get_mode(value):
        """File mode converter"""
        mode = 0
        flags = (stat.S_IFDIR,
                 stat.S_IRUSR, stat.S_IWUSR, stat.S_IXUSR,
                 stat.S_IRGRP, stat.S_IWGRP, stat.S_IXGRP,
                 stat.S_IROTH, stat.S_IWOTH, stat.S_IXOTH)
        for idx, flag in enumerate(value):
            if flag != '-':
                mode |= flags[idx]
        return mode

    def stat(self, name):  # pylint: disable=too-many-locals
        # get mode
        mode = 0
        entries = []
        try:
            self.ftp_client.retrlines('LIST {}'.format(self.get_path(name)),
                                      entries.append)
        except Error as ex:
            raise FileNotFoundError(name) from ex
        if len(entries) == 1:
            mode, ignore, user, group, size, month, day, year, name = entries[0].split()  # pylint: disable=unused-variable
            mode = self.get_mode(mode)
        else:
            for entry in entries:
                mode, ignore, user, group, size, month, day, year, name = entry.split()  # pylint: disable=unused-variable
                if name == '.':
                    mode = self.get_mode(mode)
                    break
        # get size
        try:
            size = self.ftp_client.size(self.get_path(name))
        except:  # pylint: disable=bare-except
            size = 0
        # get time
        try:
            _ret, mdtm = self.ftp_client.voidcmd('MDTM {}'.format(self.get_path(name))).split()
            mtime = round(mktime(time.strptime(mdtm, '%Y%m%d%H%M%S')))
        except:  # pylint: disable=bare-except
            mtime = 0
        return os.stat_result((
            mode,   # st_mode
            0,      # st_ino
            0,      # st_dev
            0,      # st_nlink
            0,      # st_uid
            0,      # st_gid
            size,   # st_size
            mtime,  # st_atime
            mtime,  # st_mtime
            mtime   # st_ctime
        ))

    def set_time(self, name, atime, mtime):
        pass

    def set_mode(self, name, mode=0x400):
        pass

    def unlink(self, name):
        self.ftp_client.delete(self.get_path(name))



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS scheduler.include module

This module is used for Pyramid integration.
"""

import atexit
import logging
import os.path
import re
import sys

from pyramid.events import subscriber
from pyramid.interfaces import IApplicationCreated
from pyramid.settings import asbool
from zope.interface.interfaces import ComponentLookupError

from pyams_scheduler.interfaces import MANAGE_SCHEDULER_PERMISSION, MANAGE_TASKS_PERMISSION, \
    SCHEDULER_AUTH_KEY, SCHEDULER_CLIENTS_KEY, SCHEDULER_HANDLER_KEY, SCHEDULER_MANAGER_ROLE, \
    SCHEDULER_NAME, SCHEDULER_STARTER_KEY, TASKS_MANAGER_ROLE
from pyams_scheduler.process import SchedulerMessageHandler, SchedulerProcess
from pyams_security.interfaces import ADMIN_USER_ID, SYSTEM_ADMIN_ROLE
from pyams_security.interfaces.base import MANAGE_ROLES_PERMISSION, ROLE_ID
from pyams_site.interfaces import PYAMS_APPLICATION_DEFAULT_NAME, PYAMS_APPLICATION_SETTINGS_KEY
from pyams_utils.protocol.tcp import is_port_in_use
from pyams_utils.registry import get_pyramid_registry, set_local_registry
from pyams_utils.zodb import get_connection_from_settings
from pyams_zmq.process import process_exit_func


__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


LOGGER = logging.getLogger('PyAMS (scheduler)')


def include_package(config):
    """Pyramid package include"""

    # add translations
    config.add_translation_dirs('pyams_scheduler:locales')

    # add scheduler permissions and roles
    config.register_permission({
        'id': MANAGE_SCHEDULER_PERMISSION,
        'title': _("Manage scheduler properties")
    })
    config.register_permission({
        'id': MANAGE_TASKS_PERMISSION,
        'title': _("Manage scheduler tasks")
    })

    # upgrade system manager roles
    config.upgrade_role(SYSTEM_ADMIN_ROLE,
                        permissions={
                            MANAGE_SCHEDULER_PERMISSION,
                            MANAGE_TASKS_PERMISSION
                        })

    # register new roles
    config.register_role({
        'id': SCHEDULER_MANAGER_ROLE,
        'title': _("Scheduler manager (role)"),
        'permissions': {
            MANAGE_ROLES_PERMISSION,
            MANAGE_SCHEDULER_PERMISSION,
            MANAGE_TASKS_PERMISSION
        },
        'managers': {
            ADMIN_USER_ID,
            ROLE_ID.format(SYSTEM_ADMIN_ROLE)
        }
    })
    config.register_role({
        'id': TASKS_MANAGER_ROLE,
        'title': _("Tasks manager (role)"),
        'permissions': {
            MANAGE_TASKS_PERMISSION
        },
        'managers': {
            ADMIN_USER_ID,
            ROLE_ID.format(SYSTEM_ADMIN_ROLE),
            ROLE_ID.format(SCHEDULER_MANAGER_ROLE)
        }
    })

    try:
        import pyams_zmi  # pylint: disable=import-outside-toplevel,unused-import
    except ImportError:
        config.scan(ignore=[re.compile(r'pyams_scheduler\..*\.zmi\.?.*').search])
    else:
        config.scan()


@subscriber(IApplicationCreated)
def handle_new_application(event):  # pylint: disable=unused-argument,too-many-locals,too-many-branches
    """Start scheduler process when application is created"""

    # Check for PyAMS command line script
    cmdline = os.path.split(sys.argv[0])[-1]
    if cmdline.startswith('pyams_'):
        return

    registry = get_pyramid_registry()
    settings = registry.settings
    start_handler = asbool(settings.get(SCHEDULER_STARTER_KEY, False))
    if not start_handler:
        return

    # check if port is available
    handler_address = settings.get(SCHEDULER_HANDLER_KEY, '127.0.0.1:5555')
    hostname, port = handler_address.split(':')
    if is_port_in_use(int(port), hostname):
        LOGGER.warning("Scheduler port already used, aborting...")
        return

    # get database connection
    connection = get_connection_from_settings(settings)
    root = connection.root()
    # get application
    application_name = settings.get(PYAMS_APPLICATION_SETTINGS_KEY,
                                    PYAMS_APPLICATION_DEFAULT_NAME)
    application = root.get(application_name)
    if application is None:
        return

    process = None
    sm = application.getSiteManager()  # pylint: disable=invalid-name
    set_local_registry(sm)
    try:
        scheduler_util = sm.get(SCHEDULER_NAME)
        if scheduler_util is None:
            return
        try:
            zodb_name = scheduler_util.zodb_name
        except ComponentLookupError:
            pass
        else:
            # create scheduler process
            process = SchedulerProcess(handler_address,
                                       SchedulerMessageHandler,
                                       settings.get(SCHEDULER_AUTH_KEY),
                                       settings.get(SCHEDULER_CLIENTS_KEY),
                                       registry)
            # load tasks
            for task in scheduler_util.values():
                if not task.is_runnable():
                    continue
                trigger = task.get_trigger()
                LOGGER.debug("Adding scheduler job for task '{0.name}'".format(task))
                process.scheduler.add_job(task, trigger,
                                          id=str(task.internal_id),
                                          name=task.name,
                                          kwargs={
                                              'zodb_name': zodb_name,
                                              'registry': registry
                                          })
            # start process
            LOGGER.info("Starting tasks scheduler {0!r}...".format(process))
            process.start()
            if process.is_alive():
                atexit.register(process_exit_func, process=process)
                LOGGER.info("Started tasks scheduler with PID {0}.".format(process.pid))
    finally:
        if process and not process.is_alive():
            process.terminate()
            process.join()
        set_local_registry(None)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.task.zmi.notification module

This module provides components used to manage tasks notifications.
"""

from pyramid.view import view_config
from zope.interface import Interface
from zope.intid import IIntIds

from pyams_form.ajax import ajax_form_config
from pyams_form.field import Fields
from pyams_form.interfaces.form import IAJAXFormRenderer
from pyams_layer.interfaces import IPyAMSLayer
from pyams_pagelet.pagelet import pagelet_config
from pyams_scheduler.interfaces import IScheduler, ITask, MANAGE_TASKS_PERMISSION
from pyams_scheduler.interfaces.task import IMailNotification, ITaskNotification, \
    ITaskNotificationContainer
from pyams_scheduler.task.zmi import TaskBaseFormMixin
from pyams_scheduler.zmi import SchedulerTasksTable
from pyams_skin.interfaces.viewlet import IContentPrefixViewletManager
from pyams_skin.viewlet.menu import MenuItem
from pyams_table.interfaces import IColumn, IValues
from pyams_utils.adapter import ContextRequestViewAdapter, adapter_config
from pyams_utils.registry import get_utility
from pyams_utils.traversing import get_parent
from pyams_utils.url import absolute_url
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import AdminModalAddForm, AdminModalDisplayForm, AdminModalEditForm
from pyams_zmi.helper.container import delete_container_element, switch_element_attribute
from pyams_zmi.helper.event import get_json_table_row_add_callback, \
    get_json_table_row_refresh_callback
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.interfaces.table import ITableElementEditor
from pyams_zmi.interfaces.viewlet import IContextAddingsViewletManager
from pyams_zmi.table import ActionColumn, IconColumn, InnerTableAdminView, NameColumn, Table, \
    TableElementEditor, TrashColumn, VisibilityColumn


__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


@adapter_config(name='notifications',
                required=(IScheduler, IAdminLayer, SchedulerTasksTable),
                provides=IColumn)
class TaskNotificationsColumn(ActionColumn):
    """Task notifications column"""

    hint = _("Task notifications")
    icon_class = 'far fa-envelope'

    href = 'notifications.html'

    weight = 35


@pagelet_config(name='notifications.html',
                context=ITask, layer=IPyAMSLayer,
                permission=MANAGE_TASKS_PERMISSION)
class TaskNotificationsView(TaskBaseFormMixin, AdminModalDisplayForm):
    """Task notifications view"""


#
# Task notifications table
#

class TaskNotificationsTable(Table):
    """Task notifications table"""

    @property
    def data_attributes(self):
        attributes = super().data_attributes
        container = ITaskNotificationContainer(self.context)
        attributes['table'].update({
            'data-ams-location': absolute_url(container, self.request),
            'data-ams-order': '1,asc'
        })
        return attributes

    display_if_empty = True


@adapter_config(required=(ITask, IAdminLayer, TaskNotificationsTable),
                provides=IValues)
class TaskNotificationsTableValues(ContextRequestViewAdapter):
    """Task notifications table values adapter"""

    @property
    def values(self):
        """Task notifications table values getter"""
        yield from ITaskNotificationContainer(self.context).values()


@adapter_config(name='enabled',
                required=(ITask, IAdminLayer, TaskNotificationsTable),
                provides=IColumn)
class TaskNotificationsTableEnabledColumn(VisibilityColumn):
    """Task notifications table enabled column"""

    hint = _("Click icon to enable or disable notification")

    object_data = {
        'ams-modules': 'container',
        'ams-update-target': 'switch-active-notification.json',
        'ams-attribute-name': 'enabled',
        'ams-icon-on': 'far fa-bell',
        'ams-icon-off': 'far fa-bell-slash'
    }

    def get_icon_class(self, item):
        """Icon class getter"""
        return 'far fa-bell' if item.enabled else 'far fa-bell-slash'


@view_config(name='switch-active-notification.json',
             context=ITaskNotificationContainer, request_type=IPyAMSLayer,
             renderer='json', xhr=True)
def switch_active_notification(request):
    """Switch active notification"""
    return switch_element_attribute(request)


@adapter_config(name='label',
                required=(ITask, IAdminLayer, TaskNotificationsTable),
                provides=IColumn)
class TaskNotificationsTableNameColumn(NameColumn):
    """Task notifications table label column"""

    i18n_header = _("Target")


@adapter_config(name='send-errors-only',
                required=(ITask, IAdminLayer, TaskNotificationsTable),
                provides=IColumn)
class TaskNotificationsTableErrorsColumn(IconColumn):
    """Task notifications table errors-only column"""

    icon_class = 'fas fa-bug'
    hint = _("Send errors reports only")

    weight = 50

    def get_icon_class(self, item):
        """Icon class getter"""
        if item.report_errors_only:
            return self.icon_class
        return '{} text-secondary'.format(self.icon_class)

    def get_icon_hint(self, item):
        """Icon hint getter"""
        hint = self.hint if item.report_errors_only else _("Send all reports")
        return self.request.localizer.translate(hint)


@adapter_config(name='send-empty',
                required=(ITask, IAdminLayer, TaskNotificationsTable),
                provides=IColumn)
class TaskNotificationsTableEmptyColumn(IconColumn):
    """Task notifications table empty reports column"""

    icon_class = 'far fa-file'
    hint = _("Send empty reports")

    weight = 55

    def get_icon_class(self, item):
        """Icon class getter"""
        if item.send_empty_reports:
            return self.icon_class
        return '{} text-secondary'.format(self.icon_class)

    def get_icon_hint(self, item):
        """Icon hint getter"""
        hint = self.hint if item.send_empty_reports else _("Don't send empty reports")
        return self.request.localizer.translate(hint)


@adapter_config(name='trash',
                required=(ITask, IAdminLayer, TaskNotificationsTable),
                provides=IColumn)
class TaskNotificationsTableTrashColumn(TrashColumn):
    """Task notifications table trash column"""


@view_config(name='delete-element.json',
             context=ITaskNotificationContainer, request_type=IPyAMSLayer,
             renderer='json', xhr=True,
             permission=MANAGE_TASKS_PERMISSION)
def delete_task_notification(request):
    """Delete task notification"""
    return delete_container_element(request)


#
# Task notifications table view
#

@viewlet_config(name='task-notifications-table',
                context=ITask, layer=IAdminLayer, view=TaskNotificationsView,
                manager=IContentPrefixViewletManager, weight=10)
class TaskNotificationsTableView(InnerTableAdminView):
    """Task notifications table view"""

    title = _("Task notifications")
    table_class = TaskNotificationsTable
    table_label = _("List of defined task notifications")


#
# Base notifications add and edit forms
#

class TaskNotificationAddForm(TaskBaseFormMixin, AdminModalAddForm):
    """Base task notification add form"""

    legend = _("New notification properties")

    def add(self, obj):
        intids = get_utility(IIntIds)
        container = ITaskNotificationContainer(self.context)
        container[hex(intids.register(obj))[2:]] = obj


@adapter_config(required=(ITask, IAdminLayer, TaskNotificationAddForm),
                provides=IAJAXFormRenderer)
class TaskNotificationAddFormRenderer(ContextRequestViewAdapter):
    """Task notification add form AJAX renderer"""

    def render(self, changes):
        """AJAX result renderer"""
        if not changes:
            return None
        return {
            'callbacks': [
                get_json_table_row_add_callback(self.context, self.request,
                                                TaskNotificationsTable, changes)
            ]
        }


@adapter_config(required=(ITaskNotification, IAdminLayer, Interface),
                provides=ITableElementEditor)
class TaskNotificationEditor(TableElementEditor):
    """Task notification editor adapter"""


class TaskNotificationEditForm(TaskBaseFormMixin, AdminModalEditForm):
    """Base task notification edit form"""

    legend = _("Notification properties")


@adapter_config(required=(ITaskNotification, IAdminLayer, TaskNotificationEditForm),
                provides=IAJAXFormRenderer)
class TaskNotificationEditFormRenderer(ContextRequestViewAdapter):
    """Task notification edit form renderer"""

    def render(self, changes):
        """AJAX result renderer"""
        if changes is None:
            return None
        task = get_parent(self.context, ITask)
        return {
            'status': 'success',
            'message': self.request.localizer.translate(self.view.success_message),
            'callbacks': [
                get_json_table_row_refresh_callback(task, self.request,
                                                    TaskNotificationsTable, self.context)
            ]
        }


#
# Mail notifications views
#

@viewlet_config(name='add-mail-notification.menu',
                context=ITask, layer=IAdminLayer, view=TaskNotificationsTable,
                manager=IContextAddingsViewletManager, weight=10,
                permission=MANAGE_TASKS_PERMISSION)
class MailNotificationAddMenu(MenuItem):
    """Mail notification add menu"""

    label = _("Add mail notification...")
    href = 'add-mail-notification.html'
    modal_target = True

    def get_href(self):
        """Link URL getter"""
        return absolute_url(self.context, self.request, self.href)


@ajax_form_config(name='add-mail-notification.html',
                  context=ITask, layer=IPyAMSLayer,
                  permission=MANAGE_TASKS_PERMISSION)
class MailNotificationAddForm(TaskNotificationAddForm):
    """Mail notification add form"""

    fields = Fields(IMailNotification).omit('__name__', '__parent__', 'enabled')
    content_factory = IMailNotification


@ajax_form_config(name='properties.html',
                  context=ITaskNotification, layer=IPyAMSLayer,
                  permission=MANAGE_TASKS_PERMISSION)
class MailNotificationEditForm(TaskNotificationEditForm):
    """Mail notification edit form"""

    fields = Fields(IMailNotification).omit('__name__', '__parent__', 'enabled')



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.zmi.task.rest module

This module defines REST API caller task management interface.
"""

from urllib import parse

from pyramid.events import subscriber
from zope.interface import Interface, Invalid, alsoProvides, implementer

from pyams_form.ajax import ajax_form_config
from pyams_form.browser.checkbox import SingleCheckBoxFieldWidget
from pyams_form.field import Fields
from pyams_form.group import GroupManager
from pyams_form.interfaces.form import IDataExtractedEvent, IForm, IGroup, IInnerTabForm
from pyams_form.subform import InnerAddForm, InnerEditForm
from pyams_layer.interfaces import IPyAMSLayer
from pyams_scheduler.interfaces import IScheduler, MANAGE_TASKS_PERMISSION
from pyams_scheduler.interfaces.task.rest import IRESTCallerTask, IRESTCallerTaskInfo
from pyams_scheduler.task.rest import RESTCallerTask
from pyams_scheduler.zmi import SchedulerTasksTable
from pyams_scheduler.task.zmi import BaseTaskAddForm, BaseTaskEditForm
from pyams_skin.interfaces.viewlet import IHelpViewletManager
from pyams_skin.viewlet.help import AlertMessage
from pyams_skin.viewlet.menu import MenuItem
from pyams_utils.adapter import adapter_config
from pyams_utils.interfaces.data import IObjectData
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import FormGroupChecker
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.interfaces.viewlet import IContextAddingsViewletManager


__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


class IRESTTaskForm(IForm):
    """REST caller forms common interface"""


@implementer(IRESTTaskForm)
class RESTTaskFormInfo(GroupManager):
    """REST API caller task add form info"""

    title = _("HTTP/REST API settings")
    fields = Fields(IRESTCallerTaskInfo).select('base_url', 'service', 'params', 'verify_ssl',
                                                'connection_timeout', 'ok_status',
                                                'allow_redirects')
    fields['verify_ssl'].widget_factory = SingleCheckBoxFieldWidget
    fields['allow_redirects'].widget_factory = SingleCheckBoxFieldWidget

    def update_widgets(self, prefix=None):
        """Widgets update"""
        super().update_widgets(prefix)  # pylint: disable=no-member
        params = self.widgets.get('params')  # pylint: disable=no-member
        if params is not None:
            params.add_class('height-100')
            params.widget_css_class = "editor height-200px"
            params.object_data = {
                'ams-filename': 'params.json'
            }
            alsoProvides(params, IObjectData)


@subscriber(IDataExtractedEvent, form_selector=RESTTaskFormInfo)
def extract_rest_task_info(event):
    """Extract REST task info"""
    data = event.data
    if 'base_url' in data:
        parsed_url = parse.urlparse(data['base_url'])
        if not (parsed_url.scheme and parsed_url.netloc):
            event.form.widgets.errors += (Invalid(_("Missing base target URI!")), )


@adapter_config(name='rest-auth.group',
                required=(IScheduler, IAdminLayer, IRESTTaskForm),
                provides=IGroup)
@adapter_config(name='rest-auth.group',
                required=(IRESTCallerTask, IAdminLayer, IRESTTaskForm),
                provides=IGroup)
class RESTTaskFormAuthInfo(FormGroupChecker):
    """REST task form auth info"""

    legend = _("Authentication")
    fields = Fields(IRESTCallerTaskInfo).select('authenticate', 'username', 'password')

    weight = 10

    def update_widgets(self, prefix=None):
        super().update_widgets(prefix)
        password = self.widgets.get('password')
        if password is not None:
            password.autocomplete = 'new-password'


@subscriber(IDataExtractedEvent, form_selector=RESTTaskFormAuthInfo)
def extract_rest_auth_info(event):
    """Extract REST task auth info"""
    data = event.data
    authenticate = data.get('authenticate')
    if authenticate and not data.get('username'):
        event.form.widgets.errors += (Invalid(_("Username and password are required to use "
                                                "authentication!")), )


@adapter_config(name='rest-proxy.group',
                required=(IScheduler, IAdminLayer, IRESTTaskForm),
                provides=IGroup)
@adapter_config(name='rest-proxy.group',
                required=(IRESTCallerTask, IAdminLayer, IRESTTaskForm),
                provides=IGroup)
class RESTTaskFormProxyInfo(FormGroupChecker):
    """REST task form proxy info"""

    legend = _("Proxy settings")
    fields = Fields(IRESTCallerTaskInfo).select('use_proxy', 'proxy_server', 'proxy_port',
                                                'proxy_username', 'proxy_password')

    weight = 20

    def update_widgets(self, prefix=None):
        super().update_widgets(prefix)
        password = self.widgets.get('proxy_password')
        if password is not None:
            password.autocomplete = 'new-password'


@subscriber(IDataExtractedEvent, form_selector=RESTTaskFormProxyInfo)
def extract_rest_proxy_info(event):
    """Extract REST task proxy info"""
    data = event.data
    use_proxy = data.get('use_proxy')
    if use_proxy and not data.get('proxy_server'):
        event.form.widgets.errors += (Invalid(_("Proxy access defined without proxy server!")), )


@adapter_config(name='rest-jwt-authority.group',
                required=(IScheduler, IAdminLayer, IRESTTaskForm),
                provides=IGroup)
@adapter_config(name='rest-jwt-authority.group',
                required=(IRESTCallerTask, IAdminLayer, IRESTTaskForm),
                provides=IGroup)
class RESTTaskFormJWTInfo(FormGroupChecker):
    """REST task form JWT info"""

    fields = Fields(IRESTCallerTaskInfo).select('use_jwt_authority', 'jwt_authority_url',
                                                'jwt_token_service', 'jwt_token_attribute',
                                                'jwt_use_proxy')
    fields['jwt_use_proxy'].widget_factory = SingleCheckBoxFieldWidget

    weight = 30


@viewlet_config(name='rest-jwt-authority.help',
                context=Interface, layer=IAdminLayer, view=RESTTaskFormJWTInfo,
                manager=IHelpViewletManager, weight=10)
class RESTTaskFormJWTHelp(AlertMessage):
    """REST task form JWT help"""

    css_class = 'mb-1 p-2'

    _message = _("If this service require a JWT token, you can define the authentication "
                 "authority which will be used to get new access tokens.")


@subscriber(IDataExtractedEvent, form_selector=RESTTaskFormJWTInfo)
def extract_rest_jwt_info(event):
    """Extract REST task JWT info"""
    data = event.data
    use_jwt_authority = data.get('use_jwt_authority')
    if use_jwt_authority and not data.get('jwt_authority_url'):
        event.form.widgets.errors += (Invalid(_("JWT authority location is required to enable "
                                                "JWT authentication!")), )


#
# REST API task add forms
#

@viewlet_config(name='add-rest-task.menu',
                context=IScheduler, layer=IAdminLayer, view=SchedulerTasksTable,
                manager=IContextAddingsViewletManager, weight=20,
                permission=MANAGE_TASKS_PERMISSION)
class RESTTaskAddMenu(MenuItem):
    """REST task add menu"""

    label = _("Add HTTP/REST API caller task...")
    href = 'add-rest-task.html'
    modal_target = True


@ajax_form_config(name='add-rest-task.html', context=IScheduler, layer=IPyAMSLayer,
                  permission=MANAGE_TASKS_PERMISSION)
class RESTTaskAddForm(BaseTaskAddForm):
    """REST task add form"""

    content_factory = IRESTCallerTask
    content_label = RESTCallerTask.label


@adapter_config(name='rest-task-info.form',
                required=(IScheduler, IAdminLayer, RESTTaskAddForm),
                provides=IInnerTabForm)
class RESTTaskAddFormInfo(RESTTaskFormInfo, InnerAddForm):
    """REST task add form info"""


#
# REST task edit forms
#

@ajax_form_config(name='properties.html', context=IRESTCallerTask, layer=IPyAMSLayer,
                  permission=MANAGE_TASKS_PERMISSION)
class RESTTaskEditForm(BaseTaskEditForm):
    """REST task edit form"""


@adapter_config(name='rest-task-info.form',
                required=(IRESTCallerTask, IAdminLayer, RESTTaskEditForm),
                provides=IInnerTabForm)
class RESTTaskEditFormInfo(RESTTaskFormInfo, InnerEditForm):
    """REST task edit form info"""



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.process module

This module defines the main scheduler base classes, to handle scheduler process and main
tasks management threads.
"""

__docformat__ = 'restructuredtext'

import logging
from datetime import datetime
from threading import Thread

from apscheduler.jobstores.memory import MemoryJobStore
from apscheduler.schedulers.background import BackgroundScheduler
from pyramid.config import Configurator
from pyramid.threadlocal import manager
from zope.interface import implementer

from pyams_scheduler.interfaces import ISchedulerProcess, ITask, SCHEDULER_NAME
from pyams_scheduler.trigger import ImmediateTaskTrigger
from pyams_site.interfaces import PYAMS_APPLICATION_DEFAULT_NAME, PYAMS_APPLICATION_SETTINGS_KEY
from pyams_utils.zodb import ZODBConnection
from pyams_zmq.handler import ZMQMessageHandler
from pyams_zmq.process import ZMQProcess


LOGGER = logging.getLogger('PyAMS (scheduler)')


class BaseTaskThread(Thread):
    """Base task management thread class"""

    def __init__(self, process, settings):
        Thread.__init__(self)
        self.process = process
        if ITask.providedBy(settings):
            scheduler = settings.__parent__
            self.settings = {
                'zodb_name': scheduler.zodb_name,
                'task_name': settings.__name__,
                'job_id': settings.internal_id
            }
        else:
            self.settings = settings

    def run(self):
        """Thread start method

        Subclasses may override this method, but must call this super()
        method to correctly initialize ZCA hook in the thread.
        """
        registry = self.process.registry
        manager.push({'registry': registry, 'request': None})
        config = Configurator(registry=registry)
        config.hook_zca()

    def _get_connection(self):
        """ZODB connection getter"""
        return ZODBConnection(name=self.settings.get('zodb_name'))


class TaskResettingThread(BaseTaskThread):
    """Task resetting thread

    Task reset is run in another thread, so that:
    - other transactions applied on updated tasks are visible
    - ØMQ request returns immediately to calling process
    """

    def run(self):
        super().run()
        LOGGER.debug("Starting task resetting thread...")
        settings = self.settings
        job_id = settings.get('job_id')
        if job_id is None:
            return
        job_id = str(job_id)
        LOGGER.debug("Loading ZODB connection...")
        with self._get_connection() as root:
            LOGGER.debug("Loaded ZODB root {0!r}".format(root))
            try:
                registry = self.process.registry
                application_name = registry.settings.get(PYAMS_APPLICATION_SETTINGS_KEY,
                                                         PYAMS_APPLICATION_DEFAULT_NAME)
                application = root.get(application_name)
                LOGGER.debug("Loaded application {0!r}".format(application))
                sm = application.getSiteManager()  # pylint: disable=invalid-name
                scheduler_util = sm.get(SCHEDULER_NAME)
                LOGGER.debug("Loaded scheduler utility {0!r}".format(scheduler_util))
                scheduler = self.process.scheduler
                LOGGER.debug("Removing job '{0}'".format(job_id))
                job = scheduler.get_job(job_id)
                if job is not None:
                    LOGGER.debug("Loaded job {0!r} ({0.id!r})".format(job))
                    scheduler.remove_job(job.id)
                LOGGER.debug("Loading scheduler task '{0}'".format(
                    settings.get('task_name').lower()))
                task = scheduler_util.get(settings.get('task_name').lower())
                LOGGER.debug("Loaded scheduler task {0!r}".format(task))
                if (task is not None) and task.is_runnable():
                    trigger = task.get_trigger()
                    LOGGER.debug("Getting task trigger {0!r}".format(trigger))
                    LOGGER.debug("Adding new job to scheduler {0!r}".format(scheduler))
                    scheduler.add_job(task, trigger,
                                      id=str(task.internal_id),
                                      name=task.name,
                                      kwargs={
                                          'zodb_name': scheduler_util.zodb_name,
                                          'registry': registry
                                      })
                    LOGGER.debug("Added job")
            except:  # pylint: disable=bare-except
                LOGGER.exception("An exception occurred:")


class TaskRemoverThread(BaseTaskThread):
    """Task remover thread"""

    def run(self):
        super().run()
        LOGGER.debug("Starting task remover thread...")
        settings = self.settings
        job_id = settings.get('job_id')
        if job_id is None:
            return
        job_id = str(job_id)
        LOGGER.debug("Loading ZODB connection...")
        with self._get_connection() as root:
            LOGGER.debug("Loaded ZODB root {0!r}".format(root))
            try:
                registry = self.process.registry
                application_name = registry.settings.get(PYAMS_APPLICATION_SETTINGS_KEY,
                                                         PYAMS_APPLICATION_DEFAULT_NAME)
                application = root.get(application_name)
                LOGGER.debug("Loaded application {0!r}".format(application))
                sm = application.getSiteManager()  # pylint: disable=invalid-name
                scheduler_util = sm.get(SCHEDULER_NAME)
                LOGGER.debug("Loaded scheduler utility {0!r}".format(scheduler_util))
                scheduler = self.process.scheduler
                LOGGER.debug("Removing job '{0}'".format(job_id))
                job = scheduler.get_job(job_id)
                if job is not None:
                    LOGGER.debug("Loaded job {0!r} ({0.id!r})".format(job))
                    scheduler.remove_job(job.id)
                LOGGER.debug("Removed job")
            except:  # pylint: disable=bare-except
                LOGGER.exception("An exception occurred:")


class TaskRunnerThread(BaseTaskThread):
    """Task immediate runner thread"""

    def run(self):
        super().run()
        LOGGER.debug("Starting task runner thread...")
        settings = self.settings
        job_id = settings.get('job_id')
        if job_id is None:
            return
        LOGGER.debug("Loading ZODB connection...")
        with self._get_connection() as root:
            LOGGER.debug("Loaded ZODB root {0!r}".format(root))
            try:
                registry = self.process.registry
                application_name = registry.settings.get(PYAMS_APPLICATION_SETTINGS_KEY,
                                                         PYAMS_APPLICATION_DEFAULT_NAME)
                application = root.get(application_name)
                LOGGER.debug("Loaded application {0!r}".format(application))
                sm = application.getSiteManager()  # pylint: disable=invalid-name
                scheduler_util = sm.get(SCHEDULER_NAME)
                LOGGER.debug("Loaded scheduler utility {0!r}".format(scheduler_util))
                scheduler = self.process.scheduler
                LOGGER.debug("Loading scheduler task '{0}'".format(
                    settings.get('task_name').lower()))
                task = scheduler_util.get(settings.get('task_name').lower())
                LOGGER.debug("Loaded scheduler task {0!r}".format(task))
                if task is not None:
                    trigger = ImmediateTaskTrigger()
                    LOGGER.debug("Getting task trigger {0!r}".format(trigger))
                    LOGGER.debug("Adding new job to scheduler {0!r}".format(scheduler))
                    scheduler.add_job(task, trigger,
                                      id='{0.internal_id}::{1}'.format(
                                          task, datetime.utcnow().isoformat()),
                                      name=task.name,
                                      kwargs={
                                          'zodb_name': scheduler_util.zodb_name,
                                          'registry': self.process.registry,
                                          'run_immediate': True
                                      })
                    LOGGER.debug("Added job")
            except:  # pylint: disable=bare-except
                LOGGER.exception("An exception occurred:")


class SchedulerHandler:
    """Scheduler handler"""

    process = None

    def test(self, settings):  # pylint: disable=unused-argument
        """Scheduler handler test"""
        messages = [
            'OK - Tasks scheduler ready to handle requests.',
            '{0} currently running jobs'.format(len(self.process.scheduler.get_jobs()))
        ]
        return [200, '\n'.join(messages)]

    def get_jobs(self, settings):  # pylint: disable=unused-argument
        """Getter of currently running jobs"""
        scheduler = self.process.scheduler
        return [200, [{
            'id': job.id,
            'name': job.name,
            'trigger': '{0!s}'.format(job.trigger),
            'next_run': job.next_run_time.timestamp()
        } for job in scheduler.get_jobs()]]

    def reset_task(self, settings):
        """Reset task with given properties"""
        TaskResettingThread(self.process, settings).start()
        return [200, 'OK']

    def remove_task(self, settings):
        """Remove task with given properties"""
        TaskRemoverThread(self.process, settings).start()
        return [200, 'OK']

    def run_task(self, settings):
        """Run task with given properties"""
        TaskRunnerThread(self.process, settings).start()
        return [200, 'OK']


class SchedulerMessageHandler(ZMQMessageHandler):
    """ØMQ scheduler messages handler"""

    handler = SchedulerHandler


@implementer(ISchedulerProcess)
class SchedulerProcess(ZMQProcess):
    """ØMQ tasks scheduler process"""

    def __init__(self, zmq_address, handler, auth, clients, registry):
        # pylint: disable=too-many-arguments
        super().__init__(zmq_address, handler, auth, clients, registry)
        self.scheduler = BackgroundScheduler()
        self.jobstore = MemoryJobStore()

    def run(self):
        if self.scheduler is not None:
            self.scheduler.add_jobstore(self.jobstore, 'default')
            self.scheduler.start()
        ZMQProcess.run(self)
